/*
 *                            COPYRIGHT
 *
 *  PCB, interactive printed circuit board design
 *  Copyright (C) 1994,1995,1996,1997,1998,1999 Thomas Nau
 *
 *  This module, output.c, was written and is Copyright (C) 1999 harry eaton
 *  portions of code taken from main.c by Thomas Nau
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Contact addresses for paper mail and Email:
 *  Thomas Nau, Schlehenweg 15, 88471 Baustetten, Germany
 *  Thomas.Nau@rz.uni-ulm.de
 *
 */

#define SCROLL_TIME 25

#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <memory.h>

#include "global.h"

#include "action.h"
#include "buffer.h"
#include "create.h"
#include "control.h"
#include "crosshair.h"
#include "data.h"
#include "draw.h"
#include "error.h"
#include "file.h"
#include "library.h"
#include "menu.h"
#include "misc.h"
#include "mymem.h"
#include "output.h"
#include "remove.h"
#include "set.h"

#include <X11/cursorfont.h> 
#include <X11/Shell.h> 
#include <X11/Xutil.h>
#include <X11/Xatom.h> 
#include <X11/Xaw/Command.h> 
#include <X11/Xaw/Form.h> 
#include <X11/Xaw/Label.h> 
#include <X11/Xaw/Panner.h>
#include <X11/Xaw/Porthole.h>
#include <X11/Xaw/Scrollbar.h> 
#include <X11/Xaw/Simple.h> 

/* ----------------------------------------------------------------------
 * Function prototype declarations
 */
static void UpdateExposed(XEvent *);
static void CB_ScrollX(XtPointer, XtIntervalId *);
static void CB_ScrollY(XtPointer, XtIntervalId *);
static void CB_Panner(Widget, XtPointer, XtPointer);
static Boolean ActiveDrag(void);

static XtIntervalId SxID = 0, SyID = 0;
static Position LastX, LastY;

/* ----------------------------------------------------------------------
 * sets the size of the drawing area
 * is called by MapNotify and ConfigureNotify events
 */
void GetSizeOfDrawingArea(void)
{
	Dimension	width, height;

	XtVaGetValues(Output.Porthole,
		XtNwidth, &width,
		XtNheight, &height,
		NULL);

		/* adjust the size of the statusline */
	XtVaSetValues(Output.StatusLine, XtNwidth, width, NULL);
	
		/* get a new pixmap if necessary */
	if (width != Output.Width || height != Output.Height)
	{
			/* can't call NewPixmap() because the sizes aren't right yet ?*/ 
		ReleaseOffscreenPixmap();
		Offmask = XCreatePixmap(Dpy, Output.OutputWindow, width, height, 1);
		Offscreen = XCreatePixmap(Dpy, Output.OutputWindow, width, height,
			DefaultDepthOfScreen(XtScreen(Output.Output)));
		if (VALID_PIXMAP(Offscreen))
			XFillRectangle(Dpy, Offscreen, Output.bgGC, 0, 0, width, height);
		if (VALID_PIXMAP(Offmask) && Output.pmGC == (GC) 0)
			Output.pmGC = XCreateGC(Dpy, Offmask, 0, NULL);
		if (!VALID_PIXMAP(Offmask))
			Message("Damn! Can't get the bitmap!!\n");
		ScaleOutput(width, height);
			/* this is the porthole size */
		Output.Width = width;
		Output.Height = height;
		UpdateAll();
	}
}
void ScaleOutput(Dimension width, Dimension height)
{
	Dimension	cwidth, scale;
	Position	xmore, ymore;
	Position	maxw = TO_SCREEN(PCB->MaxWidth),
			maxh = TO_SCREEN(PCB->MaxHeight);

		/* compute the output size so the panner
		 * has constant aspect ratio and minimum
		 * size
		 */
	ymore = MAX(0, MAX(height, 3 * width / 4) - maxh);
	xmore = 4 * (maxh + ymore) / 3 - maxw;
	if (xmore < 0)
	{
		ymore -= 3 * xmore / 4;
		xmore = 0;
	}

	XtMakeResizeRequest(Output.Output, maxw + xmore, maxh + ymore, NULL, NULL);
	Output.cw = maxw + xmore;
	Output.ch = maxh + ymore;
	XtVaGetValues(Output.Control, XtNwidth, &cwidth, NULL);
	cwidth -= 10;
	scale = (100 * cwidth) / (maxw + xmore);
	XtVaSetValues(Output.panner, XtNcanvasWidth, Output.cw,
		XtNcanvasHeight, Output.ch,
		XtNdefaultScale, scale,
		XtNsliderHeight, height,
		XtNsliderWidth, width,
		NULL);
}

static void CB_ScrollX(XtPointer unused, XtIntervalId *time)
{
	SxID = 0;
	HideCrosshair(False);
	CenterDisplay(LastX, 0, True);
	MoveCrosshairRelative(TO_PCB(LastX), 0);
	RestoreCrosshair(False);
	SxID = XtAppAddTimeOut(Context, SCROLL_TIME, CB_ScrollX, NULL);
}

static void CB_ScrollY(XtPointer unused, XtIntervalId *time)
{
	SyID = 0;
	HideCrosshair(False);
	CenterDisplay(0, LastY, True);
	MoveCrosshairRelative(0, TO_PCB(LastY));
	RestoreCrosshair(False);
	SyID = XtAppAddTimeOut(Context, SCROLL_TIME, CB_ScrollY, NULL);
}

void Pan(Position X, Position Y, Boolean Scroll, Boolean Update)
{
	static Position x, y;
	Dimension	scale;
	Pixmap		draw_here;

	if (X < 0)
		X = 0;
	else if (X > Output.cw - Output.Width)
		X = Output.cw - Output.Width;
	if (Y < 0)
		Y = 0;
	else if (Y > Output.ch - Output.Height)
		Y = Output.ch - Output.Height;
	Xorig = -X;
	Yorig = -Y;
	render = True;

		/* ignore the no - change case */
	if (Xorig == x && Yorig == y)
		return;
		/* scroll the pixmap */
	if (Scroll)
	{
		Position	sx, sy;

		if (VALID_PIXMAP(Offscreen))
		{
			draw_here = Offscreen;
			sx = sy = 0;
		}
		else
		{
			draw_here = Output.OutputWindow;
			sx = x;
			sy = y;
		}
		XCopyArea(Dpy, draw_here, draw_here, Output.bgGC, sx, sy,
			Output.Width, Output.Height,
			sx + Xorig - x, sy + Yorig - y);
			/* do same for Offmask ? */
		if (VALID_PIXMAP(Offmask))
			XCopyArea(Dpy, Offmask, Offmask, Output.pmGC, 0, 0,
				Output.Width, Output.Height,
				Xorig - x, Yorig - y);
		if (y != Yorig)
			XFillRectangle(Dpy, draw_here, Output.bgGC, 0,
				(Yorig - y > 0) ? sy : sy + Output.Height
				+ Yorig - y + sy, Output.Width,
				abs(y - Yorig));
		if (x != Xorig)
			XFillRectangle(Dpy, draw_here, Output.bgGC,
				(Xorig - x > 0) ? sx : sx + Output.Width
				+ Xorig - x + sx, 0, abs(x - Xorig),
				Output.Height);
	}
	x = Xorig;
	y = Yorig;

	XtVaSetValues(Output.Output, XtNx, x, XtNy, y, NULL);
	Output.OffsetX = -Xorig;
	Output.OffsetY = -Yorig;
	if (Update)
	{
		XtVaGetValues(Output.panner, XtNdefaultScale, &scale, NULL);
		XtVaSetValues(Output.panner, XtNsliderX, -x,
			XtNsliderY, -y, NULL);
	}
}

/* ---------------------------------------------------------------------- 
 * handles all events from porthole window
 */
void PortholeEvent(Widget W,
	XtPointer ClientData, XEvent *Event, Boolean *Flag)
{
	if (Event->type == ConfigureNotify)
		GetSizeOfDrawingArea();
} 

static void UpdateExposed(XEvent *Event)
{
	XRectangle		clipBox;
	XEvent	others;
	Region myRegion = XCreateRegion();

		/* make sure all geometry changes have taken effect */
	XSync(Dpy, False);
	if (Event)
		XtAddExposureToRegion(Event, myRegion);
	while (XtAppPending(Context))
	{
		XtAppNextEvent(Context, &others);
			/* any future re-configuration obviates prior updates */
		if (others.type != Expose ||
		    ((XExposeEvent *) &others)->window != Output.OutputWindow)
			XtDispatchEvent(&others);
		else
			XtAddExposureToRegion(&others, myRegion);
	}
		/* pixmap and screen have different clip regions */
	if (VALID_PIXMAP(Offscreen))
	{
		XClipBox(myRegion, &clipBox);
		XOffsetRegion(myRegion, Xorig, Yorig);
	}
		/* clip drawing to only exposed region */
	XSetRegion(Dpy, Output.fgGC, myRegion);
	XSetRegion(Dpy, Output.bgGC, myRegion);
	XSetRegion(Dpy, Output.pmGC, myRegion);
	XSetRegion(Dpy, Output.GridGC, myRegion);
	RedrawOutput();
	XSetRegion(Dpy, Output.fgGC, FullRegion);
	XSetRegion(Dpy, Output.bgGC, FullRegion);
	XSetRegion(Dpy, Output.pmGC, FullRegion);
	XSetRegion(Dpy, Output.GridGC, FullRegion);
	XDestroyRegion(myRegion);
		/* if output is drawn in pixmap, copy it */
	if (VALID_PIXMAP(Offscreen))
	{
		HideCrosshair(True);
/*
		XSync(Dpy, False);
*/
		if (!XCopyArea(Dpy, Offscreen, Output.OutputWindow,
			Output.fgGC,
			clipBox.x -Output.OffsetX, clipBox.y -Output.OffsetY,
			clipBox.width, clipBox.height,
			clipBox.x, clipBox.y))
		{
				/* shouldn't be possible to fail, but... */
			ReleaseOffscreenPixmap();
			RedrawOutput();
			Message("hace: problem with XCopyArea!\n");
		}
		RestoreCrosshair(True);
	}
}

static Boolean ActiveDrag(void)
{
	Boolean active = False;
	
	switch (Settings.Mode)
	{
		case POLYGON_MODE:
			if (Crosshair.AttachedLine.State != STATE_FIRST)
				active = True;
			break;

		case ARC_MODE:
			if (Crosshair.AttachedBox.State != STATE_FIRST)
				active = True;
			break;

		case LINE_MODE:
			if (Crosshair.AttachedLine.State != STATE_FIRST)
				active = True;
			break;

		case COPY_MODE:
		case MOVE_MODE:
		case INSERTPOINT_MODE:
			if (Crosshair.AttachedObject.Type != NO_TYPE)
				active = True;
			break;
		default:
			break;
	}

	if (Crosshair.AttachedBox.State == STATE_SECOND)
		active = True;

	return(active);
}

/* ---------------------------------------------------------------------- 
 * handles all events from output window
 * the handling of MapNotify and ConfigureNotify could be done much
 * easier with X11R5 and later but...
 *
 * I also had to set a flag to indicate a valid 'enter window' because
 * modal windows may generate 'LeaveNotify' events without 'EnterNotify'.
 * This behavior disturbs the crosshair-visibility management
 */
void OutputEvent(Widget W,
	XtPointer ClientData, XEvent *Event, Boolean *Flag)
{
	int			ksym;
	static	Boolean		has_entered = False;

	switch(Event->type)
	{
		case Expose:	/* expose event means redraw */
		{
			UpdateExposed(Event);
			break;
		}

		case EnterNotify:			/* enter output area */
			has_entered = True;
				/* stop scrolling if necessary */
			if (SxID)
			{
				XtRemoveTimeOut(SxID);
				SxID = 0;
				break;
			}
			if (SyID)
			{
				XtRemoveTimeOut(SyID);
				SyID = 0;
				break;
			}
			RestoreCrosshair(False);
			break;

		case LeaveNotify:			/* leave output area */
				/* see also the description at start of this function */
			if (has_entered)
			{
				Position x, y;

					/* if actively drawing, start scrolling */
				if (ActiveDrag())
				{
					x =((XLeaveWindowEvent *) Event)->x - Output.OffsetX;
					y =((XLeaveWindowEvent *) Event)->y - Output.OffsetY;
					if (x < 0)
					{
						LastX = - Settings.Grid;
						SxID = XtAppAddTimeOut(Context, SCROLL_TIME, CB_ScrollX, NULL);
					}
					else if (x >= Output.Width)
					{
						LastX = Settings.Grid;
						SxID = XtAppAddTimeOut(Context, SCROLL_TIME, CB_ScrollX, NULL);
					}
					else if (y < 0)
					{
						LastY = -Settings.Grid;
						SyID = XtAppAddTimeOut(Context, SCROLL_TIME, CB_ScrollY, NULL);
					}
					else if (y >= Output.Height)
					{
						LastY = Settings.Grid;
						SyID = XtAppAddTimeOut(Context, SCROLL_TIME, CB_ScrollY, NULL);
					}
				}
				else
					HideCrosshair(False);
			}
			has_entered = False;	
			break;
			
		case MapNotify:				/* get initial size of drawing area */
		case ConfigureNotify:		/* get position of origin */
			GetSizeOfDrawingArea();
			break;

		case MotionNotify:
				/* the damn pop-up menu gives a leave notify */
				/* but no enter notify if the mouse is released */
				/* outside the menu window */
				/* in fact *NO* event is generated to report that */
				/* the menu is gone and we're back in the output */
				/* window ... sigh */
				/* of course movement here suggest we're back */
			if (!has_entered)
			{
				RestoreCrosshair(False);
				has_entered = True;
			}
			EventMoveCrosshair((XMotionEvent *) Event);
			break;

		case KeyPress:
		case KeyRelease:
			ksym = XKeycodeToKeysym(Dpy, ((XKeyEvent *) Event)->keycode, 0);
			if (ksym == XK_Shift_R || ksym == XK_Shift_L)
			{
				HideCrosshair(False);
				AdjustAttachedObjects();
				RestoreCrosshair(True);
			}
			break;
		default:
			break;
	}
} 

static void CB_Panner(Widget W, XtPointer ClientData, XtPointer CallData)
{
	XawPannerReport	*report = (XawPannerReport *) CallData; 

	Pan(report->slider_x, report->slider_y, True, False);
}

Widget InitOutputPanner(Widget Parent, Widget Top, Widget Left)
{
	Widget	panner;

	panner = XtVaCreateManagedWidget("panner", pannerWidgetClass,
		Parent,
		XtNfromVert, Top,
		XtNfromHoriz, Left,
		NULL);
	Output.panner = panner;
	XtAddCallback(panner, XtNreportCallback, CB_Panner, NULL);
	return(panner);
}

	
